import drjit as dr
import pytest

#  pkgs = ["drjit.cuda", "drjit.cuda.ad",
#          "drjit.llvm", "drjit.llvm.ad"]

pkgs = ["drjit.llvm", "drjit.llvm.ad"]
pkgs_ad = ["drjit.llvm.ad"]

def get_class(name):
    """Resolve a package+class name into the corresponding type"""
    if 'cuda' in name:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    elif 'llvm' in name:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')

    name = name.split('.')
    value = __import__(".".join(name[:-1]))
    for item in name[1:]:
        value = getattr(value, item)

    return value

class Checker:
    """
    Compares a Tensor indexing operation against a NumPy reference
    and asserts if there is a mismatch.
    """
    def __init__(self, shape, tensor_type):
        np = pytest.importorskip("numpy")

        self.shape = shape
        size = np.prod(shape)
        self.array_n = np.arange(size, dtype=np.uint32).reshape(shape)
        self.array_e = tensor_type(dr.arange(tensor_type.Array, size), shape)

    def __getitem__(self, args):
        np = pytest.importorskip("numpy")

        ref_n = self.array_n.__getitem__(args)
        ref_e = self.array_e.__getitem__(args)
        assert ref_n.shape == ref_e.shape
        assert np.all(ref_n.ravel() == ref_e.array.numpy())


@pytest.mark.parametrize("pkg", pkgs)
def test01_slice_index(pkg):
  t = get_class(pkg + ".TensorXu")
  c = Checker((10,), t)
  c[:]
  c[3]
  c[1:5]
  c[-5]

  c = Checker((10, 20), t)
  c[:]
  c[5, 0]
  c[5, 0:2]
  c[:, 5]
  c[5, :]
  c[:, :]
  c[1:3, 2:7:2]
  c[8:2:-1, 7:0:-1]
  c[0:0, 0:0]


@pytest.mark.parametrize("pkg", pkgs)
def test02_slice_ellipsis(pkg):
    t = get_class(pkg + ".TensorXu")
    c = Checker((10, 20, 30, 40), t)

    c[...]
    c[1, ...]
    c[..., 1]
    c[4, ..., 3]
    c[0, 1:3, ..., 3]


@pytest.mark.parametrize("pkg", pkgs)
def test03_slice_append_dim(pkg):
    t = get_class(pkg + ".TensorXu")
    c = Checker((10, 20, 30, 40), t)

    c[None]
    c[..., None]
    c[1, None, ...]
    c[..., None, 1, None]
    c[None, 4, ..., 3, None]


@pytest.mark.parametrize("pkg", pkgs)
def test04_broadcasting(pkg):
    np = pytest.importorskip("numpy")

    t = get_class(pkg + ".TensorXu")
    for i in range(1, 5):
        for j in range(1, 5):
            for k in range(1, 5):
                shape = [i, j, k]
                for l in range(len(shape)):
                    shape_2 = list(shape)
                    shape_2[l] = 1
                    array_n1 = np.arange(np.prod(shape),   dtype=np.uint32).reshape(shape)
                    array_n2 = np.arange(np.prod(shape_2), dtype=np.uint32).reshape(shape_2)

                    array_e1 = t(dr.arange(t.Index, np.prod(shape)),   shape)
                    array_e2 = t(dr.arange(t.Index, np.prod(shape_2)), shape_2)

                    out_n = array_n1 + array_n2
                    out_e = array_e1 + array_e2

                    assert out_n.shape == out_e.shape
                    assert np.all(out_n.ravel() == out_e.array.numpy())
                    assert np.all((array_n1 * 2).ravel() == (array_e1 * 2).array.numpy())

    with pytest.raises(Exception) as e:
        a = dr.full(t, 1, (3, 3))
        b = dr.full(t, 1, (2, 2))
        c = a + b
    e.match('incompatible tensor shapes for dimension')


@pytest.mark.parametrize("pkg", pkgs)
def test05_initialization_casting(pkg):
    tu = get_class(pkg + ".TensorXu")
    tf = get_class(pkg + ".TensorXf")
    tf64 = get_class(pkg + ".TensorXf")

    t0 = dr.full(tu, 1, (2, 3, 4))
    t1 = dr.full(tf, 2, (2, 3, 4))
    t2 = dr.zeros(tf64, (2, 3, 4))

    assert dr.shape(t0) == (2, 3, 4)

    t3 = t0 + t1 + t2
    assert type(t3) is tf64

    assert t3.shape == (2, 3, 4)
    assert t3.array == dr.full(t3.Array, 3, 2*3*4)

    t3[:, 1, :] = 12
    assert t3[:, 0, :] == 3
    assert t3[:, 1, :] == 12


@pytest.mark.parametrize("pkg", pkgs_ad)
def test05_ad(pkg):
    f = get_class(pkg + ".TensorXf")
    z0 = dr.full(f, 1, (2, 3, 4))
    assert not dr.grad_enabled(z0)
    dr.enable_grad(z0)
    assert dr.grad_enabled(z0)
    assert not dr.grad_enabled(dr.detach(z0))
    assert dr.ravel(z0) is z0.array

    z1 = z0 + z0
    dr.backward(z1)
    g = dr.grad(z0)
    assert g.shape == (2, 3, 4)
    assert len(g.array) == 2*3*4
    assert g.array == 2


@pytest.mark.parametrize("pkg", pkgs)
def test06_numpy_conversion(pkg):
    np = pytest.importorskip("numpy")

    f = get_class(pkg + ".TensorXf")

    value = f(dr.arange(f.Array, 2*3*4), (2, 3, 4))
    value_np = value.numpy()
    assert value_np.shape == (2, 3, 4)
    assert np.all(value_np.ravel() == value.array.numpy())

    value_2 = f(value_np)
    assert value.shape == value_2.shape
    assert value.array == value_2.array

    value_np = np.ones((1,1,1,1))
    value_3 = f(value_np)
    assert value_np.shape == value_3.shape
    assert np.all(value_np == value_3.array)


@pytest.mark.parametrize("pkg", pkgs)
def test07_jax_conversion(pkg):
    jax = pytest.importorskip("jax")
    f = get_class(pkg + ".TensorXf")

    value = f(dr.arange(f.Array, 2*3*4), (2, 3, 4))
    value_jax = value.jax()
    assert value_jax.shape == (2, 3, 4)
    assert jax.numpy.all(value_jax.ravel() == value.array.jax())

    value_2 = f(value_jax)
    assert value.shape == value_2.shape
    assert value.array == value_2.array


@pytest.mark.parametrize("pkg", pkgs)
def test08_pytorch_conversion(pkg):
    torch = pytest.importorskip("torch")
    f = get_class(pkg + ".TensorXf")

    value = f(dr.arange(f.Array, 2*3*4), (2, 3, 4))
    value_torch = value.torch()
    assert value_torch.shape == (2, 3, 4)
    assert torch.all(value_torch.flatten() == value.array.torch())

    value_2 = f(value_torch)
    assert value.shape == value_2.shape
    assert value.array == value_2.array


@pytest.mark.parametrize("pkg", pkgs)
def test09_tensorflow_conversion(pkg):
    tf = pytest.importorskip("tensorflow")
    f = get_class(pkg + ".TensorXf")
    tf.constant(0)

    value = f(dr.arange(f.Array, 2*3*4), (2, 3, 4))
    value_tf = value.tf()
    assert value_tf.shape == (2, 3, 4)
    assert tf.reduce_all(tf.equal(tf.reshape(value_tf, (2*3*4,)), value.array.tf()))

    value_2 = f(value_tf)
    assert value.shape == value_2.shape
    assert value.array == value_2.array


@pytest.mark.parametrize("pkg", pkgs)
def test10_tensorflow_arithmetic(pkg):
    t = get_class(pkg + ".TensorXf")
    f = get_class(pkg + ".Float32")

    tt = t([1, 2, 3, 4, 5, 6], [2, 3])
    ff = f(2.0)

    assert ff * tt == tt * ff
    assert ff * tt == t([2, 4, 6, 8, 10, 12], [2, 3])


class PowerOfTwo(dr.CustomOp):
    def eval(self, value):
        self.value = value
        return value * value

    def forward(self):
        grad_in = self.grad_in('value')
        self.set_grad_out(2.0 * self.value * grad_in)

    def backward(self):
        grad_out = self.grad_out()
        self.set_grad_in('value', 2.0 * self.value * grad_out)

    def name(self):
        return "power of two"


@pytest.mark.parametrize("pkg", ["drjit.llvm.ad", "drjit.cuda.ad"])
def test11_custom_op(pkg):
    t = get_class(pkg + ".TensorXf")
    f = get_class(pkg + ".Float32")

    tt = t([1, 2, 3, 4, 5, 6], [2, 3])
    dr.enable_grad(tt)

    tt2 = dr.custom(PowerOfTwo, tt)

    dr.set_grad(tt2, 1.0)
    dr.enqueue(dr.ADMode.Backward, tt2)
    dr.traverse(f, dr.ADMode.Backward)

    assert dr.grad(tt).array == [2.0, 4.0, 6.0, 8.0, 10.0, 12.0]


@pytest.mark.parametrize("pkg", pkgs_ad)
def test12_select(pkg):
    for tp in [get_class(pkg + ".TensorXf"), get_class(pkg + ".TensorXu")]:
        initial = tp([1, 2, 3, 4], shape=(4, 1))

        next = initial + 10
        valid = initial >= 2.5
        assert type(valid) == dr.mask_t(initial)

        result = dr.select(valid, next, initial)
        assert type(result) == tp

        expected = tp([1, 2, 13, 14], shape=dr.shape(initial))
        assert dr.allclose(result, expected)


@pytest.mark.parametrize("pkg", pkgs)
def test13_upsampling_tensor(pkg):
    t = get_class(pkg + ".TensorXf")

    a = t([1, 2, 3, 4], shape=(2, 2))
    assert dr.allclose(dr.upsample(a, [4, 4]).array, [1, 1, 2, 2,
                                                      1, 1, 2, 2,
                                                      3, 3, 4, 4,
                                                      3, 3, 4, 4])

    b = dr.upsample(a, scale_factor=[3, 3])
    assert dr.allclose(b.array, [1, 1, 1, 2, 2, 2,
                                 1, 1, 1, 2, 2, 2,
                                 1, 1, 1, 2, 2, 2,
                                 3, 3, 3, 4, 4, 4,
                                 3, 3, 3, 4, 4, 4,
                                 3, 3, 3, 4, 4, 4])

    b = dr.upsample(a, scale_factor=[3, 1])
    assert dr.allclose(b.array, [1, 2, 1, 2, 1, 2,
                                 3, 4, 3, 4, 3, 4])

    b = dr.upsample(a, scale_factor=[3])
    assert dr.allclose(b.array, [1, 2, 1, 2, 1, 2,
                                 3, 4, 3, 4, 3, 4])

    b = dr.upsample(a, scale_factor=[1, 3])
    assert dr.allclose(b.array, [1, 1, 1, 2, 2, 2,
                                 3, 3, 3, 4, 4, 4])

    a = t([1, 2, 3, 2, 3, 4, 3, 4, 5, 4, 5, 6], shape=(2, 2, 3))
    assert dr.allclose(dr.upsample(a, [4, 4]).array, [1, 2, 3, 1, 2, 3, 2, 3, 4, 2, 3, 4,
                                                      1, 2, 3, 1, 2, 3, 2, 3, 4, 2, 3, 4,
                                                      3, 4, 5, 3, 4, 5, 4, 5, 6, 4, 5, 6,
                                                      3, 4, 5, 3, 4, 5, 4, 5, 6, 4, 5, 6])

    a = t([1, 2, 3, 2, 3, 4, 3, 4, 5, 4, 5, 6], shape=(2, 2, 3))
    assert dr.allclose(dr.upsample(a, [4, 4, 3]).array, [1, 2, 3, 1, 2, 3, 2, 3, 4, 2, 3, 4,
                                                         1, 2, 3, 1, 2, 3, 2, 3, 4, 2, 3, 4,
                                                         3, 4, 5, 3, 4, 5, 4, 5, 6, 4, 5, 6,
                                                         3, 4, 5, 3, 4, 5, 4, 5, 6, 4, 5, 6])

    a = t([1, 2, 3, 4, 5, 6, 7, 8], shape=(2, 2, 2))
    assert dr.allclose(dr.upsample(a, [4, 4, 4]).array, [1, 1, 2, 2, 1, 1, 2, 2, 3, 3, 4, 4, 3, 3, 4, 4,
                                                         1, 1, 2, 2, 1, 1, 2, 2, 3, 3, 4, 4, 3, 3, 4, 4,
                                                         5, 5, 6, 6, 5, 5, 6, 6, 7, 7, 8, 8, 7, 7, 8, 8,
                                                         5, 5, 6, 6, 5, 5, 6, 6, 7, 7, 8, 8, 7, 7, 8, 8])

    with pytest.raises(TypeError) as ei:
        dr.upsample(a.array, [4])
    assert "unsupported input type" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.upsample(a, shape=[4], scale_factor=[4])
    assert "shape and scale_factor" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.upsample(a, shape=3)
    assert "unsupported shape type" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.upsample(a, shape=[2, 2, 2, 2])
    assert "invalid shape size" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.upsample(a, shape=[2, 2, 2.5])
    assert "must contain integer values" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.upsample(a, shape=[1, 1, 1])
    assert "must be larger" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.upsample(a, shape=[3, 3, 3])
    assert "must be multiples" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.upsample(a, scale_factor=3)
    assert "unsupported scale_factor type" in str(ei.value)

    with pytest.raises(TypeError) as ei:
        dr.upsample(a, scale_factor=[2, 2, 0])
    assert "must be greater than 0" in str(ei.value)


@pytest.mark.parametrize("pkg", pkgs)
def test14_upsampling_texture(pkg):
    t = get_class(pkg + ".TensorXf")
    tex_t = get_class(pkg + ".Texture2f")

    a = tex_t(t([1, 2, 3, 4], shape=(2, 2, 1)), filter_mode=dr.FilterMode.Nearest)
    b = dr.upsample(a, shape=[4, 4])
    assert dr.allclose(b.tensor().array, [1, 1, 2, 2,
                                          1, 1, 2, 2,
                                          3, 3, 4, 4,
                                          3, 3, 4, 4])

    a = tex_t(t([1, 2, 3, 4], shape=(2, 2, 1)), filter_mode=dr.FilterMode.Nearest)
    b = dr.upsample(a, shape=[3, 3])
    assert dr.allclose(b.tensor().array, [1, 2, 2,
                                          3, 4, 4,
                                          3, 4, 4])

    a = tex_t(t([1, 2, 3, 4], shape=(2, 2, 1)), filter_mode=dr.FilterMode.Linear)
    b = dr.upsample(a, shape=[4, 4])
    assert dr.allclose(b.tensor().array, [1.0, 1.25, 1.75, 2.0,
                                          1.5, 1.75, 2.25, 2.5,
                                          2.5, 2.75, 3.25, 3.5,
                                          3.0, 3.25, 3.75, 4.0])

    a = tex_t(t([1, 2, 3, 4], shape=(2, 2, 1)), filter_mode=dr.FilterMode.Linear)
    b = dr.upsample(a, shape=[3, 3])
    assert dr.allclose(b.tensor().array, [1.0, 1.5, 2.0,
                                          2.0, 2.5, 3.0,
                                          3.0, 3.5, 4.0])

    a = tex_t(t([1, 1, 5, 2, 2, 6, 3, 3, 7, 4, 4, 8], shape=(2, 2, 3)), filter_mode=dr.FilterMode.Linear)
    b = dr.upsample(a, shape=[3, 3])
    assert dr.allclose(b.tensor().array, [1.0, 1.0, 5.0, 1.5, 1.5, 5.5, 2.0, 2.0, 6.0,
                                          2.0, 2.0, 6.0, 2.5, 2.5, 6.5, 3.0, 3.0, 7.0,
                                          3.0, 3.0, 7.0, 3.5, 3.5, 7.5, 4.0, 4.0, 8.0])



import sys
import os

if sys.version_info < (3, 8):
    raise ImportError("Dr.Jit requires Python >= 3.8")

# Implementation details accessed by both C++ and Python
import drjit.detail as detail # noqa

if os.name != 'nt':
    # Use RTLD_DEEPBIND to prevent the DLL to search symbols in the global scope
    old_flags = sys.getdlopenflags()
    new_flags = os.RTLD_LAZY | os.RTLD_LOCAL
    if sys.platform != 'darwin':
        new_flags |= os.RTLD_DEEPBIND
    sys.setdlopenflags(new_flags)
    del new_flags

# Native extension defining low-level arrays
import drjit.drjit_ext as drjit_ext  # noqa

if os.name != 'nt':
    sys.setdlopenflags(old_flags)
    del old_flags

# Routing functionality (type promotion, broadcasting, etc.)
import drjit.router as router  # noqa

# Generic fallback implementations of array operations
import drjit.generic as generic  # noqa

# Type traits analogous to the ones provided in C++
import drjit.traits as traits  # noqa

# Math library and const
import drjit.const as const  # noqa

# Matrix-related functions
import drjit.matrix as matrix # noqa

# Tensor-related functions
import drjit.tensor as tensor # noqa

# Install routing functions in ArrayBase and global scope
self = vars()
base = self['ArrayBase']
for k, v in router.__dict__.items():
    if k.startswith('_') or (k[0].isupper() and not k == 'CustomOp'):
        continue
    if k.startswith('op_'):
        setattr(base, '__' + k[3:] + '__', v)
    else:
        self[k] = v

# Install generic array functions in ArrayBase
for k, v in generic.__dict__.items():
    if k.startswith('_') or k[0].isupper():
        continue
    if k.startswith('op_'):
        setattr(base, '__' + k[3:] + '__', v)
    else:
        setattr(base, k, v)


# Install type traits in global scope
for k, v in traits.__dict__.items():
    if k.startswith('_') or k[0].isupper():
        continue
    self[k] = v


# Install constants in global scope
for k, v in const.__dict__.items():
    if k.startswith('_'):
        continue
    self[k] = v


# Install matrix-related functions in global scope
for k, v in matrix.__dict__.items():
    if k.startswith('_') or k[0].isupper():
        continue
    self[k] = v


# Install tensor-related functions
for k, v in tensor.__dict__.items():
    if k.startswith('_') or k[0].isupper():
        continue
    self[k] = v

del sys, os
del k, v, self, base, generic, router, matrix, tensor, traits, const, drjit_ext



import drjit as _dr
import sys
import inspect
from collections.abc import Mapping, Sequence

VAR_TYPE_NAME = [
    "Void",    "Bool",  "Int8",  "UInt8", "Int16",
    "UInt16",  "Int",   "UInt",  "Int64", "UInt64", "Pointer",
    "Float16", "Float", "Float64"
]

VAR_TYPE_SUFFIX = [
    "???", "b", "i8",  "u8",  "i16", "u16", "i", "u",
    "i64", "u64", "p", "f16", "f", "f64"
]


def array_name(prefix, vt, shape, scalar):
    """
    Determines the name of an array (e.g. Float32, ArrayXf32, etc.). This
    function is used when arrays are created during initialization of the Dr.Jit
    extension module, and during implicit type promotion where array types are
    determined dynamically.

    Parameter ``prefix`` (``str``):
        Array flavor prefix (Array/Matrix/Complex/Quaternion)

    Parameter ``vt`` (``drjit.VarType``):
        Underlying scalar type (e.g. ``VarType.Int32``) of the desired array

    Parameter ``shape`` (``Tuple[int]``):
        Size per dimension

    Parameter ``scalar`` (``bool``):
        Arrays in the ``drjit.scalar.*`` module use a different depth
        convention, which is indicated via this parameter.
    """

    if not scalar and not prefix == 'Tensor':
        shape = shape[:-1]
    if prefix == "Matrix":
        if vt != _dr.VarType.Bool:
            shape = shape[1:]
        else:
            prefix = "Array"

    if len(shape) == 0 and prefix != "Tensor":
        return VAR_TYPE_NAME[int(vt)]

    return "%s%s%s" % (
        prefix,
        "".join(repr(s) if s != _dr.Dynamic else "X" for s in shape),
        VAR_TYPE_SUFFIX[int(vt)]
    )


def array_from_dlpack(t, capsule):
    descr = _dr.detail.from_dlpack(capsule)

    device_type = descr["device_type"]
    data = descr["data"]
    dtype = descr["dtype"]
    shape = descr["shape"]
    ndim = len(shape)
    strides = descr["strides"]

    if strides is None:
        tmp = 1
        strides = [0] * ndim
        for i in reversed(range(ndim)):
            strides[i] = tmp
            tmp *= shape[i]

    if t.IsCUDA and device_type != 2:
        raise _dr.Exception("Cannot create an Dr.Jit GPU array from a "
                              "DLPack CPU tensor!")
    elif not t.IsCUDA and device_type != 1:
        raise _dr.Exception("Cannot create an Dr.Jit CPU array from a "
                              "DLPack GPU tensor!")

    if dtype != t.Type:
        raise _dr.Exception("Incompatible type!")

    shape_target = list(reversed(_dr.shape(t())))
    if len(shape_target) != ndim:
        raise _dr.Exception("Incompatible dimension!")
    for i in range(ndim):
        if shape_target[i] != shape[i] and shape_target[i] != 0:
            raise _dr.Exception("Incompatible shape!")

    value = t
    while issubclass(value.Value, _dr.ArrayBase):
        value = value.Value

    descr["consume"](capsule)
    data = value.map_(data, _dr.prod(shape), descr["release"])

    def load(t, i, offset):
        size = shape[-1 - i]
        stride = strides[-1 - i]

        if i == ndim - 1:
            if type(offset) is int and stride == 1:
                return data
            else:
                i = _dr.arange(_dr.int32_array_t(t), size)
                return t.gather_(data, offset + stride * i, True, False)
        else:
            result = t()
            for j in range(size):
                result[j] = load(t.Value, i + 1, offset + stride * j)
            return result

    return load(t, 0, 0)


def sub_len(o):
    ol = len(o[0])
    try:
        for i in range(1, len(o)):
            if len(o[i]) != ol:
                return -1
    except Exception:
        return -1
    return ol


def array_init(self, args):
    """
    This generic initialization routine initializes an arbitrary Dr.Jit array
    from a variable-length argument list (which could be a scalar broadcast, a
    component list, or a NumPy/PyTorch/Tensorflow array..)
    """
    n = len(args)
    if n == 0:
        return

    size = self.Size
    value_type = self.Value
    dynamic = size == _dr.Dynamic
    err = None

    try:
        if n == 1:
            o = args[0]
            t = type(o)
            mod = t.__module__
            name = t.__name__
            is_array = issubclass(t, _dr.ArrayBase)
            is_static_array = is_array and not o.Size == _dr.Dynamic
            is_sequence = issubclass(t, Sequence) and not issubclass(t, str)

            # Matrix initialization from nested list
            if is_sequence and self.IsMatrix and \
                len(o) == size and sub_len(o) == size:
                for x in range(size):
                    for y in range(size):
                        self[x, y] = value_type.Value(o[x][y])
            elif is_array or is_sequence:
                os = len(o)
                if dynamic:
                    size = os
                    self.init_(size)

                if size == 0:
                    pass
                elif size != os or value_type is t:
                    # Size mismatch!
                    if self.IsMatrix and getattr(t, "IsMatrix", False):
                        # If both are matrices, copy the top-left block
                        for x in range(size):
                            for y in range(size):
                                if x < o.Size and y < o.Size:
                                    self[x, y] = value_type.Value(o[x, y])
                                else:
                                    self[x, y] = value_type.Value(1 if x == y else 0)
                    elif self.IsMatrix and value_type is t:
                        for x in range(size):
                            self[x] = o
                    else:
                        # Otherwise, try to broadcast to all entries
                        self.broadcast_(value_type(o)
                                        if not issubclass(t, value_type)
                                        and not self.IsMatrix else o)
                else:
                    # Size matches, copy element by element
                    if self.IsJIT and getattr(t, "IsJIT", False) and \
                       self.Depth == 1 and t.Depth == 1:
                        raise _dr.Exception(
                            "Refusing to do an extremely inefficient "
                            "element-by-element array conversion from type %s "
                            "to %s. Did you forget a cast or detach operation?"
                            % (str(type(o)), str(type(self))))

                    if isinstance(o[0], value_type) or self.IsMatrix:
                        for i in range(size):
                            self.set_entry_(i, o[i])
                    else:
                        for i in range(size):
                            self.set_entry_(i, value_type(o[i]))
            elif issubclass(t, (int, float)) and (not self.IsJIT or self.Depth > 1):
                if dynamic:
                    size = 1
                    self.init_(size)
                self.broadcast_(o)
            elif issubclass(t, complex) and self.IsComplex:
                self.set_entry_(0, o.real)
                self.set_entry_(1, o.imag)
            elif mod == "numpy":
                import numpy as np

                s1 = tuple(reversed(_dr.shape(self)))
                s2 = o.shape

                # Remove unnecessary outer dimension is possible
                if s2[0] == 1 and len(s2) > 1:
                    o = o[0, ...]
                    s2 = o.shape

                if o.dtype == np.complex64:
                    s2 = (*s2, 2)
                    o = o.view(np.float32).reshape(s2)
                elif o.dtype == np.complex128:
                    s2 = (*s2, 2)
                    o = o.view(np.float64).reshape(s2)

                if o.dtype != self.Type.NumPy:
                    o = o.astype(self.Type.NumPy)

                dim1 = len(s1)
                dim2 = len(s2)

                # Numpy array might have one dimension less when initializing dynamic arrays
                if not dim1 == dim2 and not (dim1 == dim2 + 1 and self.IsDynamic):
                    raise _dr.Exception("Incompatible dimension!")
                for i in reversed(range(dim2)):
                    if s1[i] != s2[i] and s1[i] != 0:
                        raise _dr.Exception("Incompatible shape!")

                if dim1 == 0:
                    pass
                elif dim1 == 1 and self.IsDynamic:
                    o = np.ascontiguousarray(o)
                    holder = (o, o.__array_interface__["data"][0])
                    self.assign(self.load_(holder[1], s2[0]))
                else:
                    for i in range(s1[-1]):
                        if dim2 == 1 and self.IsDynamic:
                            self.set_entry_(i, value_type.Value(o[i]))
                        else:
                            self.set_entry_(i, value_type(o[..., i]))

            elif mod == "builtins" and name == "PyCapsule":
                self.assign(array_from_dlpack(type(self), o))
            elif mod == "torch":
                from torch.utils.dlpack import to_dlpack
                self.assign(array_from_dlpack(type(self), to_dlpack(o)))
            elif mod.startswith("tensorflow."):
                from tensorflow.experimental.dlpack import to_dlpack
                self.assign(array_from_dlpack(type(self), to_dlpack(o)))
            elif mod.startswith("jax.") or mod.startswith("jaxlib."):
                from jax.dlpack import to_dlpack
                self.assign(array_from_dlpack(type(self), to_dlpack(o)))
            else:
                raise _dr.Exception("Don\"t know how to create an Dr.Jit array "
                                      "from type \"%s.%s\"!" % (mod, name))
        elif n == size or dynamic:
            if dynamic:
                size = n
                self.init_(size)
            for i in range(size):
                self.set_entry_(i, value_type(args[i]))
        elif self.IsMatrix and n == self.Size * self.Size:
            tbl = [[args[i*self.Size + j] for i in range(self.Size)]
                   for j in range(self.Size)]
            array_init(self, tbl)
        else:
            raise _dr.Exception("Invalid size!")
    except Exception as e:
        err = e

    if err is not None:
        if dynamic:
            raise TypeError("%s constructor expects: arbitrarily many values "
                            "of type \"%s\", a matching list/tuple, or a NumPy/"
                            "PyTorch/TF/Jax array." % (type(self).__name__,
                                                       value_type.__name__)) from err
        else:
            raise TypeError("%s constructor expects: %s%i values "
                            "of type \"%s\", a matching list/tuple, or a NumPy/"
                            "PyTorch/TF/Jax array." % (type(self).__name__, "" if
                                                       size == 1 else "1 or ", size,
                                                       value_type.__name__)) from err


def tensor_init(tensor_type, obj):
    mod = type(obj).__module__
    if 'tensorflow' in mod:
        import tensorflow as tf
        return tensor_type(tensor_type.Array(tf.reshape(obj, [-1])), obj.shape)
    elif mod.startswith('torch'):
        shape = obj.shape if len(obj.shape) > 0 else [1]
        return tensor_type(tensor_type.Array(obj.flatten()), shape)
    elif mod.startswith(('numpy', 'jax')):
        return tensor_type(tensor_type.Array(obj.ravel()), obj.shape)
    else:
        info = getattr(obj, '__array_interface__', None)
        if info is not None:
            shape = info['shape']
            typestr = str(info['typestr'])[3:-1]
            cls = tensor_type.Array
            if typestr != tensor_type.Array.Type.NumPy:
                name = None
                for v in VAR_TYPE_NAME:
                    t = getattr(_dr.VarType, v, None)
                    if t and t.NumPy == typestr:
                        name = array_name('Array', t, [_dr.Dynamic], False)
                        break
                if name and hasattr(mod, name):
                    cls = getattr(mod, name)
                else:
                    import numpy as np
                    np_data = np.array(obj).astype(tensor_type.Array.Type.NumPy)
                    return tensor_init(tensor_type, np_data)

            data = cls.load_(info['data'][0], _dr.prod(shape))
            return tensor_type(tensor_type.Array(data), shape)
        else:
            raise TypeError("TensorXf: expect an array that implements the "
                            "array interface protocol!")

@property
def prop_x(self):
    return self[0]


@prop_x.setter
def prop_x(self, value):
    self[0] = value


@property
def prop_y(self):
    return self[1]


@prop_y.setter
def prop_y(self, value):
    self[1] = value


@property
def prop_z(self):
    return self[2]


@prop_z.setter
def prop_z(self, value):
    self[2] = value


@property
def prop_w(self):
    return self[3]


@prop_w.setter
def prop_w(self, value):
    self[3] = value


@property
def prop_xyz(self):
    return self.Imag(self[0], self[1], self[2])


@prop_xyz.setter
def prop_xyz(self, value):
    if not isinstance(value, self.Imag):
        value = self.Imag(value)
    self.x = value.x
    self.y = value.y
    self.z = value.z


def array_configure(cls, shape, type_, value):
    """Populates an Dr.Jit array class with extra type trait fields"""
    depth = 1

    cls.Value = value
    cls.Type = type_
    cls.Shape = shape
    cls.Size = shape[0]
    cls.IsDynamic = cls.Size == _dr.Dynamic or \
        getattr(value, "IsDynamic", False)

    while issubclass(value, _dr.ArrayBase):
        value = value.Value
        depth += 1

    cls.Depth = depth
    cls.Scalar = value
    cls.IsDrJit = True
    cls.IsMask = issubclass(value, bool)
    cls.IsIntegral = issubclass(value, int) and not cls.IsMask
    cls.IsFloat = issubclass(value, float)
    cls.IsArithmetic = cls.IsIntegral or cls.IsFloat
    cls.IsSigned = cls.IsFloat or "i" in VAR_TYPE_SUFFIX[int(type_)]

    mod = cls.__module__
    cls.IsScalar = mod.startswith("drjit.scalar")
    cls.IsPacket = mod.startswith("drjit.packet")
    cls.IsDiff = mod.endswith(".ad")
    cls.IsLLVM = mod.startswith("drjit.llvm")
    cls.IsCUDA = mod.startswith("drjit.cuda")
    cls.IsJIT = cls.IsLLVM or cls.IsCUDA

    name = cls.__name__
    cls.IsMatrix = "Matrix" in name
    cls.IsComplex = "Complex" in name
    cls.IsQuaternion = "Quaternion" in name
    cls.IsTensor = "Tensor" in name
    cls.IsSpecial = cls.IsMatrix or cls.IsComplex or cls.IsQuaternion
    cls.IsVector = cls.Size != _dr.Dynamic and not \
        (cls.IsPacket and cls.Depth == 1) and not cls.IsSpecial

    prefix = name
    for i, c in enumerate(name):
        if c.isdigit() or c == 'X':
            prefix = name[:i]
            break
    mask_name = prefix

    cls.Prefix = prefix

    if cls.IsSpecial:
        mask_name = 'Array'

        if cls.IsComplex:
            cls.real = prop_x
            cls.imag = prop_y
        elif cls.IsQuaternion:
            cls.real = prop_w
            cls.imag = prop_xyz
            cls.Imag = getattr(sys.modules.get(mod),
                               name.replace("Quaternion4", "Array3"))
            cls.Complex = getattr(sys.modules.get(mod),
                                  name.replace("Quaternion4", "Complex2"))

    if cls.IsTensor:
        cls.__getitem__ = _dr.detail.tensor_getitem
        cls.__setitem__ = _dr.detail.tensor_setitem

    elif (not cls.IsSpecial or cls.IsQuaternion) \
            and not cls.Size == _dr.Dynamic:
        if cls.Size > 0:
            cls.x = prop_x
        if cls.Size > 1:
            cls.y = prop_y
        if cls.Size > 2:
            cls.z = prop_z
        if cls.Size > 3:
            cls.w = prop_w

    cls.MaskType = getattr(
        sys.modules.get(mod),
        array_name(mask_name, _dr.VarType.Bool, cls.Shape, cls.IsScalar))


def _loop_process_state(value: type, in_state: list, out_state: list,
                        write: bool, in_struct: bool = False):
    '''
    This helper function is used by ``drjit.*.Loop`` to collect the set of loop
    state variables and ensure that their types stay consistent over time. It
    traverses a python object tree in ``value`` and writes state variable
    indices to ``out_state``. If provided, it performs a consistency check
    against the output of a prior call provided via ``in_state``. If ``write``
    is set to ``True``, it mutates the input value based on the information in
    ``in_state``.
    '''
    t = type(value)

    out_state.append(t)
    if in_state:
        assert len(in_state) > 0
        t_old = in_state.pop()
        if t is not t_old:
            raise _dr.Exception(
                "loop_process_state(): the type of loop state variables must "
                "remain the same throughout the loop. However, one of the "
                "supplied variables changed from type %s to %s!"
                % (t_old.__name__, t.__name__))

    if issubclass(t, tuple) or issubclass(t, list):
        for entry in value:
            _loop_process_state(entry, in_state, out_state, write, in_struct)
        return

    if _dr.is_tensor_v(t):
        _loop_process_state(value.array, in_state, out_state, write, in_struct)
    elif _dr.is_jit_v(t):
        if t.Depth > 1:
            for i in range(len(value)):
                _loop_process_state(value.entry_ref_(i), in_state,
                                    out_state, write, in_struct)
        else:
            index = value.index
            index_ad = value.index_ad if t.IsDiff else 0

            if index_ad != 0 and _dr.flag(_dr.JitFlag.LoopRecord):
                raise _dr.Exception(
                    "loop_process_state(): one of the supplied loop state variables "
                    "of type %s is attached to the AD graph (i.e., grad_enabled(..) "
                    "is true). However, propagating derivatives through multiple "
                    "iterations of a recorded loop is not supported (and never "
                    "will be). Please see the documentation on differentiating loops "
                    "for details and suggested alternatives." % t.__name__)

            if index == 0:
                raise _dr.Exception(
                    "loop_process_state(): one of the supplied loop state "
                    "variables of type %s is uninitialized!" % t.__name__)

            ad_float_precision = value.Type.Size * 8 if (t.IsDiff and t.IsFloat) else 0
            out_state.append((index, index_ad, ad_float_precision))

            if in_state:
                assert len(in_state) > 0
                index, index_ad, _ = in_state.pop()

                if write:
                    value.set_index_(index)
                    if t.IsDiff:
                        value.set_index_ad_(index_ad)
    elif _dr.is_struct_v(t):
        for k, v in t.DRJIT_STRUCT.items():
            _loop_process_state(getattr(value, k), in_state, out_state, write, True)
    elif hasattr(value, 'loop_put') or value is None:
        pass
    elif not in_struct:
        raise _dr.Exception(
            "loop_process_state(): one of the provided loop state variables "
            "was of type '%s', which is not allowed (you must use Dr.Jit "
            "arrays/structs that are managed by the JIT compiler)"
            % t.__name__)


def loop_process_state(loop, funcs, state, write):
    if len(state) == 0:
        old_state = None

        for func in funcs:
            values = func()

            if isinstance(values, Sequence) and not isinstance(values, str):
                for value in values:
                    if hasattr(value, 'loop_put'):
                        value.loop_put(loop)

            # Automatically label loop variables
            cv = inspect.getclosurevars(func)
            _dr.set_label(**cv.globals)
            _dr.set_label(**cv.nonlocals)

            del values

        assert old_state is None or len(old_state) == 0
    else:
        old_state = list(state)
        old_state.reverse()

    del loop # Keep 'loop' out of the garbage collector's reach

    state.clear()

    for func in funcs:
        _loop_process_state(func(), old_state, state, write)

    assert old_state is None or len(old_state) == 0


def slice_tensor(shape, indices, uint32):
    """
    This function takes an array shape (integer tuple) and a tuple containing
    slice indices. It returns the resulting array shape and a flattened 32-bit
    unsigned integer array containing element indices.
    """
    components = []
    ellipsis = False
    none_count = 0
    shape_offset = 0

    for v in indices:
        if v is None:
            none_count += 1

    for v in indices:
        if v is None:
            components.append(None)
            continue

        if shape_offset >= len(shape):
            raise IndexError("slice_tensor(): too many indices specified!")

        size = shape[shape_offset]

        if isinstance(v, int):
            # Simple integer index, handle wrap-around
            if v < 0:
                v += size
            if v >= size:
                raise IndexError("slice_tensor(): index %i for dimension %i is "
                                 "out of range (size = %i)!" %
                                 (v, len(components), size))
            components.append((v, v+1, 1))
        elif isinstance(v, slice):
            # Rely on Python's slice.indices() function to determine everything
            components.append(v.indices(size))
        elif _dr.is_dynamic_array_v(v) and _dr.is_integral_v(v):
            if _dr.is_signed_v(v):
                v = uint32(_dr.select(v >= 0, v, v + size))
            components.append(v)
        elif isinstance(v, Sequence) and not isinstance(v, str):
            components.append(uint32([v2 if v2 >= 0 else v2 + size for v2 in v]))
        elif v is Ellipsis:
            if ellipsis:
                raise IndexError("slice_tensor(): multiple ellipses (...) are not allowed!")
            ellipsis = True

            for j in range(len(shape) - len(indices) + none_count + 1):
                components.append((0, shape[shape_offset], 1))
                shape_offset += 1
            continue
        else:
            raise TypeError("slice_tensor(): type '%s' cannot be used to index into a tensor!",
                            type(v).__name__)
        shape_offset += 1

    # Implicit ellipsis
    for j in range(len(shape) - shape_offset):
        components.append((0, shape[shape_offset], 1))
        shape_offset += 1

    # Compute total index size
    size_out = 1
    shape_out = []
    for comp in components:
        if comp is None:
            shape_out.append(1)
        else:
            size = len(comp if isinstance(comp, uint32) else range(*comp))
            if size != 1:
                shape_out.append(size)
                size_out *= shape_out[-1]
    shape_out = tuple(shape_out)

    index_tmp = _dr.arange(uint32, size_out)
    index_out = uint32()

    if size_out > 0:
        size_out = 1
        index_out = uint32(0)
        shape_offset = len(shape)-1

        for i in reversed(range(len(components))):
            comp = components[i]
            if comp is None:
                continue
            size = len(comp if isinstance(comp, uint32) else range(*comp))
            index_next = index_tmp // size
            index_rem = index_tmp - index_next * size

            if isinstance(comp, uint32):
                index_val = _dr.gather(uint32, comp, index_rem)
            else:
                if comp[0] >= 0 and comp[2] >= 0:
                    index_val = comp[0] + comp[2] * index_rem
                else:
                    index_val = uint32(comp[0] + comp[2] * _dr.int32_array_t(index_rem)(index_rem))

            index_out += index_val * size_out
            index_tmp = index_next
            size_out *= shape[shape_offset]
            shape_offset -= 1

    return shape_out, index_out


def tensor_getitem(tensor, slice_arg):
    if not isinstance(slice_arg, tuple):
        slice_arg = (slice_arg,)
    tensor_t = type(tensor)
    shape, index = slice_tensor(tensor.shape, slice_arg, tensor_t.Index)
    return tensor_t(_dr.gather(tensor_t.Array, tensor.array, index), shape)


def tensor_setitem(tensor, slice_arg, value):
    if not isinstance(slice_arg, tuple):
        slice_arg = (slice_arg,)
    tensor_t = type(tensor)
    shape, index = slice_tensor(tensor.shape, slice_arg, tensor_t.Index)
    _dr.scatter(target=tensor.array, value=value, index=index)


def diff_vars(o, indices, check_grad_enabled=True):
    """
    Extract indices of differentiable variables, returns
    the type of the underlying differentiable array
    """

    result = None
    if _dr.depth_v(o) > 1 or (isinstance(o, Sequence) and not isinstance(o, str)):
        for i in range(len(o)):
            t = diff_vars(o[i], indices, check_grad_enabled)
            if t is not None:
                result = t
    elif isinstance(o, Mapping):
        for k, v in o.items():
            t = diff_vars(v, indices, check_grad_enabled)
            if t is not None:
                result = t
    elif _dr.is_diff_v(o) and o.IsFloat:
        if _dr.is_tensor_v(o):
            result = diff_vars(o.array, indices, check_grad_enabled)
        elif o.index_ad != 0 and (not check_grad_enabled or o.grad_enabled_()):
            indices.append(o.index_ad)
            result = type(o)
    elif _dr.is_struct_v(o):
        for k in type(o).DRJIT_STRUCT.keys():
            t = diff_vars(getattr(o, k), indices, check_grad_enabled)
            if t is not None:
                result = t
    return result


def get_args_values(f, *args, **kwargs):
    '''
    Given a function, a tuple of positional arguments and a dict of keyword
    arguments, return the full tuple of positional arguments, including default
    values and keyword arguments at the right position.

    Here is a simple example:

        def f(a, b, c=1, d=2):
            pass
        get_args_values(f, 6, 5, d=4) # returns (4, 5, 1, 4)
    '''
    argcount = f.__code__.co_argcount
    argnames = f.__code__.co_varnames[:argcount]
    defaults = f.__defaults__ if f.__defaults__ else []

    assert len(args) + len(defaults) >= argcount

    m = {n: None for n in argnames}
    for i, a in enumerate(reversed(defaults)):
        m[argnames[-(i + 1)]] = a
    for i, a in enumerate(args):
        m[argnames[i]] = a
    for k, v in kwargs.items():
        m[k] = v

    return tuple(m.values())